import json
from urllib.parse import urlencode
from aiohttp import ClientSession


class RequestClient:
    async def request(
        self,
        method: str = "GET",
        url: str = "",
        headers: dict = None,
        params: dict = None,
        data: dict = None,
    ) -> dict:
        if params:
            params = urlencode(params)
        base_headers = {"User-Agent": "Mozilla/5.0", "accept-language": "en-US,en"}
        if headers:
            base_headers.update(headers)
        async with ClientSession() as session:
            async with session.request(
                method=method,
                url=url,
                headers=base_headers,
                params=params,
                data=json.dumps(data),
            ) as resp:
                headers = resp.headers
                if headers.get("Content-Type") == "application/json; charset=UTF-8":
                    response: dict = await resp.json()
                else:
                    response = await resp.read()
        return dict(response=response, headers=headers)



import re
from typing import Dict, List, Tuple, Union
from urllib.parse import parse_qs


def extract_video_id(url: str) -> str:
    pattern = r"(?:v=|\/)([0-9A-Za-z_-]{11}).*"
    result = re.search(pattern=pattern, string=url)
    return result.group(1)


def apply_descrambler(stream_data: Dict) -> Union[List, None]:
    if "url" in stream_data:
        return None
    formats = []
    if "formats" in stream_data.keys():
        formats.extend(stream_data["formats"])
    if "adaptiveFormats" in stream_data.keys():
        formats.extend(stream_data["adaptiveFormats"])
    for data in formats:
        if "url" not in data:
            if "signatureCipher" in data:
                cipher_url = parse_qs(data["signatureCipher"])
                data["url"] = cipher_url["url"][0]
                data["s"] = cipher_url["s"][0]
        data["is_otf"] = data.get("type") == "FORMAT_STREAM_TYPE_OTF"

    return formats


def mime_type_codec(mime_type_codec: str) -> Tuple[str, List[str]]:
    pattern = r"(\w+\/\w+)\;\scodecs=\"([a-zA-Z-0-9.,\s]*)\""
    regex = re.compile(pattern)
    results = regex.search(mime_type_codec)
    mime_type, codecs = results.groups()
    return mime_type, [c.strip() for c in codecs.split(",")]



from typing import Dict, List, Optional
from .request import RequestClient
from .constants import default_clients
from .extract import apply_descrambler, extract_video_id
from .streams import Stream, StreamQuery


class Client(RequestClient):
    def __init__(self, url: str, client: str = "ANDROID") -> None:
        self.video_id = extract_video_id(url)
        self.api_key = default_clients[client]["api_key"]
        self.context = default_clients[client]["context"]
        self.base_url = "https://www.youtube.com/youtubei/v1"
        self._video_info: Optional[Dict] = None
        self._fmt_streams: Optional[List[Stream]] = None

    @property
    def base_params(self) -> Dict:
        return {"key": self.api_key, "contentCheckOk": True, "racyCheckOk": True}

    @property
    def base_data(self) -> Dict:
        return {"context": self.context}

    async def video_info(self) -> Dict:
        if self._video_info:
            return self._video_info
        endpoint = f"{self.base_url}/player"
        query = {
            "videoId": self.video_id,
        }
        query.update(self.base_params)
        headers = {"Content-Type": "application/json"}
        result = await self.request(
            method="POST",
            url=endpoint,
            params=query,
            headers=headers,
            data=self.base_data,
        )
        self._video_info = result.get("response")
        return self._video_info

    async def streaming_data(self) -> Dict:
        return (await self.video_info())["streamingData"]

    async def fmt_streams(self) -> List[Stream]:
        if self._fmt_streams:
            return self._fmt_streams
        self._fmt_streams = []
        stream_manifest = apply_descrambler(await self.streaming_data())
        video_title = await self.title
        video_author = await self.author
        for stream in stream_manifest:
            video = Stream(title=video_title, author=video_author, **stream)
            self._fmt_streams.append(video)
        return self._fmt_streams

    async def streams(self) -> StreamQuery:
        return StreamQuery(await self.fmt_streams())

    @property
    async def length(self) -> int:
        """Get the video length in seconds."""
        return int(
            (await self.video_info()).get("videoDetails", {}).get("lengthSeconds")
        )

    @property
    async def title(self) -> int:
        """Get the video title."""
        return (await self.video_info()).get("videoDetails", {}).get("title")

    @property
    async def author(self) -> int:
        """Get the video author."""
        return (
            (await self.video_info()).get("videoDetails", {}).get("author", "unknown")
        )
